@@ -1,12 +1,12 @@
 # Copyright (C) 2012-2015 The python-bitcoinlib developers
 # Copyright (C) 2018 The python-bitcointx developers
 #
-# This file is part of python-bitcoinlib.
+# This file is part of python-bitcointx.
 #
 # It is subject to the license terms in the LICENSE file found in the top-level
 # directory of this distribution.
 #
-# No part of python-bitcoinlib, including this file, may be copied, modified,
+# No part of python-bitcointx, including this file, may be copied, modified,
 # propagated, or distributed except according to the terms contained in the
 # LICENSE file.
 
@@ -18,15 +18,18 @@ Functionality to build scripts, as well as SignatureHash(). Script evaluation
 is in bitcointx.core.scripteval
 """
 
-from __future__ import absolute_import, division, print_function
-
 import struct
+import hashlib
 from io import BytesIO
 
 import bitcointx.core
 import bitcointx.core._bignum
 
-from .serialize import *
+from .serialize import VarIntSerializer, BytesSerializer, ImmutableSerializable
+
+from ..util import (
+    no_bool_use_as_property, ClassMappingDispatcher, activate_class_dispatcher
+)
 
 MAX_SCRIPT_SIZE = 10000
 MAX_SCRIPT_ELEMENT_SIZE = 520
@@ -35,11 +38,33 @@ MAX_SCRIPT_OPCODES = 201
 SIGVERSION_BASE = 0
 SIGVERSION_WITNESS_V0 = 1
 
+# (partial) comment from Bitcoin Core IsStandardTx() function:
+# Biggest 'standard' txin is a 15-of-15 P2SH multisig with compressed
+# keys (remember the 520 byte limit on redeemScript size). That works
+# out to a (15*(33+1))+3=513 byte redeemScript
+MAX_P2SH_MULTISIG_PUBKEYS = 15
+
 OPCODE_NAMES = {}
 
 _opcode_instances = []
 
 
+class ScriptCoinClassDispatcher(ClassMappingDispatcher, identity='script'):
+    ...
+
+
+class ScriptCoinClass(metaclass=ScriptCoinClassDispatcher):
+    ...
+
+
+class ScriptBitcoinClassDispatcher(ScriptCoinClassDispatcher):
+    ...
+
+
+class ScriptBitcoinClass(metaclass=ScriptBitcoinClassDispatcher):
+    ...
+
+
 class CScriptOp(int):
     """A single script opcode"""
     __slots__ = []
@@ -79,6 +104,7 @@ class CScriptOp(int):
 
         return int(self - OP_1+1)
 
+    @no_bool_use_as_property
     def is_small_int(self):
         """Return true if the op pushes a small integer to the stack"""
         if 0x51 <= self <= 0x60 or self == 0:
@@ -100,9 +126,10 @@ class CScriptOp(int):
             return _opcode_instances[n]
         except IndexError:
             assert len(_opcode_instances) == n
-            _opcode_instances.append(super(CScriptOp, cls).__new__(cls, n))
+            _opcode_instances.append(super().__new__(cls, n))
             return _opcode_instances[n]
 
+
 # Populate opcode instance table
 for n in range(0xff+1):
     CScriptOp(n)
@@ -233,6 +260,7 @@ OP_NOP1 = CScriptOp(0xb0)
 OP_NOP2 = CScriptOp(0xb1)
 OP_CHECKLOCKTIMEVERIFY = OP_NOP2
 OP_NOP3 = CScriptOp(0xb2)
+OP_CHECKSEQUENCEVERIFY = OP_NOP3
 OP_NOP4 = CScriptOp(0xb3)
 OP_NOP5 = CScriptOp(0xb4)
 OP_NOP6 = CScriptOp(0xb5)
@@ -354,6 +382,7 @@ OPCODE_NAMES.update({
     OP_NOP1: 'OP_NOP1',
     OP_NOP2: 'OP_NOP2',
     OP_CHECKLOCKTIMEVERIFY: 'OP_CHECKLOCKTIMEVERIFY',
+    OP_CHECKSEQUENCEVERIFY: 'OP_CHECKSEQUENCEVERIFY',
     OP_NOP3: 'OP_NOP3',
     OP_NOP4: 'OP_NOP4',
     OP_NOP5: 'OP_NOP5',
@@ -475,6 +504,7 @@ OPCODES_BY_NAME = {
     'OP_NOP2': OP_NOP2,
     'OP_CHECKLOCKTIMEVERIFY': OP_CHECKLOCKTIMEVERIFY,
     'OP_NOP3': OP_NOP3,
+    'OP_CHECKSEQUENCEVERIFY': OP_CHECKSEQUENCEVERIFY,
     'OP_NOP4': OP_NOP4,
     'OP_NOP5': OP_NOP5,
     'OP_NOP6': OP_NOP6,
@@ -496,6 +526,61 @@ DISABLED_OPCODES = frozenset((OP_VERIF, OP_VERNOTIF,
                               OP_LSHIFT, OP_RSHIFT))
 
 
+class DATA(bytes):
+    """A class that can be used to prevent accidental use of non-data
+    elements in the script where data elemets were expected. For example,
+    the code `CScript([var])` does not communicate to the reader if `var`
+    is expected to be just data, or a number, or an opcode.
+    with CScript([DATA(var)]), this is communicated clearly, and will
+    raise TypeError if var is not bytes or bytearray instance."""
+
+    def __new__(cls, data):
+        if not isinstance(data, (bytes, bytearray)):
+            raise TypeError(
+                'DATA can only accept bytes or bytearray instance')
+
+        return super().__new__(cls, data)
+
+
+class NUMBER(int):
+    """A class that can be used to prevent accidental use of non-numeric
+    elements in the script where nmeric elemets were expected. For example,
+    the code `CScript([var])` does not communicate to the reader if `var`
+    is expected to be just data, or a number, or an opcode.
+    with CScript([NUMBER(var)]), this is communicated clearly, and will
+    raise TypeError if var is not an instance of int class, and not
+    and instance of CScriptOp (special case needed because CScriptOp is
+    a subclas of int, but there are special OPCODE guard for it"""
+
+    def __new__(cls, num):
+        if not isinstance(num, int):
+            raise TypeError(
+                'NUMBER can only accept values that are instance of '
+                'int class (except CScriptOp)')
+
+        if isinstance(num, CScriptOp):
+            raise TypeError('NUMBER can not accept CScriptOp instance')
+
+        return super().__new__(cls, num)
+
+
+def OPCODE(op):
+    """A function that can be used to prevent accidental use of non-opcode
+    elements in the script where opcode elemets were expected. For example,
+    the code `CScript([var])` does not communicate to the reader if `var`
+    is expected to be just data, or a number, or an opcode.
+    with CScript([OPCODE(var)]), this is communicated clearly, and will
+    raise TypeError if var is not an instance of CScriptOp
+    Note that while DATA and NUMBER are classes, OPCODE cannot be a class,
+    because if the op is some subclass of the CScriptOp, result of
+    OPCODE(op) will not be the same, whereas with a function the same
+    instance is just returned."""
+    if not isinstance(op, CScriptOp):
+        raise TypeError(
+            'OPCODE can only accept instances of CScriptOp')
+    return op
+
+
 class CScriptInvalidError(Exception):
     """Base class for CScript exceptions"""
     pass
@@ -505,16 +590,16 @@ class CScriptTruncatedPushDataError(CScriptInvalidError):
     """Invalid pushdata due to truncation"""
     def __init__(self, msg, data):
         self.data = data
-        super(CScriptTruncatedPushDataError, self).__init__(msg)
+        super().__init__(msg)
 
 
-class CScript(bytes):
+class CScript(bytes, ScriptCoinClass, next_dispatch_final=True):
     """Serialized script
 
     A bytes subclass, so you can use this directly whenever bytes are accepted.
     Note that this means that indexing does *not* work - you'll get an index by
     byte rather than opcode. This format was chosen for efficiency so that the
-    general case would not require creating a lot of little CScriptOP objects.
+    general case would not require creating a lot of little CScriptOp objects.
 
     iter(script) however does iterate by opcode.
     """
@@ -523,15 +608,19 @@ class CScript(bytes):
         # Coerce other into bytes
         if isinstance(other, CScriptOp):
             other = bytes([other])
-        elif isinstance(other, (int, int)):
+        elif isinstance(other, int):
             if 0 <= other <= 16:
                 other = bytes([CScriptOp.encode_op_n(other)])
             elif other == -1:
                 other = bytes([OP_1NEGATE])
             else:
-                other = CScriptOp.encode_op_pushdata(bitcointx.core._bignum.bn2vch(other))
+                other = CScriptOp.encode_op_pushdata(
+                    bitcointx.core._bignum.bn2vch(other))
         elif isinstance(other, (bytes, bytearray)):
             other = CScriptOp.encode_op_pushdata(other)
+        else:
+            raise TypeError('type {} cannot be represented in the script'
+                            .format(type(other).__name__))
         return other
 
     def __add__(self, other):
@@ -541,7 +630,7 @@ class CScript(bytes):
 
         try:
             # bytes.__add__ always returns bytes instances unfortunately
-            return CScript(super(CScript, self).__add__(other))
+            return self.__class__(super().__add__(other))
         except TypeError:
             raise TypeError('Can not add a %r instance to a CScript' % other.__class__)
 
@@ -550,15 +639,15 @@ class CScript(bytes):
         raise NotImplementedError
 
     def __new__(cls, value=b''):
-        if isinstance(value, bytes) or isinstance(value, bytearray):
-            return super(CScript, cls).__new__(cls, value)
+        if isinstance(value, (bytes, bytearray)):
+            return super().__new__(cls, value)
         else:
             def coerce_iterable(iterable):
                 for instance in iterable:
                     yield cls.__coerce_instance(instance)
             # Annoyingly on both python2 and python3 bytes.join() always
             # returns a bytes instance even when subclassed.
-            return super(CScript, cls).__new__(cls, b''.join(coerce_iterable(value)))
+            return super().__new__(cls, b''.join(coerce_iterable(value)))
 
     def raw_iter(self):
         """Raw iteration
@@ -619,7 +708,7 @@ class CScript(bytes):
     def __iter__(self):
         """'Cooked' iteration
 
-        Returns either a CScriptOP instance, an integer, or bytes, as
+        Returns either a CScriptOp instance, an integer, or bytes, as
         appropriate.
 
         See raw_iter() if you need to distinguish the different possible
@@ -642,7 +731,7 @@ class CScript(bytes):
         # For Python3 compatibility add b before strings so testcases don't
         # need to change
         def _repr(o):
-            if isinstance(o, bytes):
+            if isinstance(o, (bytes, bytearray)):
                 return "x('%s')" % bitcointx.core.b2x(o)
             else:
                 return repr(o)
@@ -665,18 +754,33 @@ class CScript(bytes):
                 if op is not None:
                     ops.append(op)
 
-        return "CScript([%s])" % ', '.join(ops)
+        return "%s([%s])" % (self.__class__.__name__, ', '.join(ops))
 
+    @no_bool_use_as_property
     def is_p2sh(self):
         """Test if the script is a p2sh scriptPubKey
 
         Note that this test is consensus-critical.
+
+        Note also that python-bitcointx does not aim to be
+        fully consensus-compatible with current Bitcoin Core codebase
         """
         return (len(self) == 23 and
                 self[0] == OP_HASH160 and
                 self[1] == 0x14 and
                 self[22] == OP_EQUAL)
 
+    @no_bool_use_as_property
+    def is_p2pkh(self):
+        """Test if the script is a p2pkh scriptPubKey"""
+        return (len(self) == 25
+                and self[0]  == OP_DUP
+                and self[1]  == OP_HASH160
+                and self[2]  == 0x14
+                and self[23] == OP_EQUALVERIFY
+                and self[24] == OP_CHECKSIG)
+
+    @no_bool_use_as_property
     def is_witness_scriptpubkey(self):
         """Returns true if this is a scriptpubkey signaling segregated witness data.
 
@@ -704,22 +808,27 @@ class CScript(bytes):
         """Returns the witness program"""
         return self[2:]
 
+    @no_bool_use_as_property
     def is_witness_v0_keyhash(self):
         """Returns true if this is a scriptpubkey for V0 P2WPKH. """
         return len(self) == 22 and self[0:2] == b'\x00\x14'
 
+    @no_bool_use_as_property
     def is_witness_v0_nested_keyhash(self):
         """Returns true if this is a scriptSig for V0 P2WPKH embedded in P2SH. """
         return len(self) == 23 and self[0:3] == b'\x16\x00\x14'
 
+    @no_bool_use_as_property
     def is_witness_v0_scripthash(self):
         """Returns true if this is a scriptpubkey for V0 P2WSH. """
         return len(self) == 34 and self[0:2] == b'\x00\x20'
 
+    @no_bool_use_as_property
     def is_witness_v0_nested_scripthash(self):
         """Returns true if this is a scriptSig for V0 P2WSH embedded in P2SH. """
         return len(self) == 35 and self[0:3] == b'\x22\x00\x20'
 
+    @no_bool_use_as_property
     def is_push_only(self):
         """Test if the script only contains pushdata ops
 
@@ -768,11 +877,12 @@ class CScript(bytes):
             return False
         return True
 
+    @no_bool_use_as_property
     def is_unspendable(self):
         """Test if the script is provably unspendable"""
-        return (len(self) > 0 and
-                self[0] == OP_RETURN)
+        return (len(self) > 0 and self[0] == OP_RETURN) or len(self) > MAX_SCRIPT_SIZE
 
+    @no_bool_use_as_property
     def is_valid(self):
         """Return True if the script is valid, False otherwise
 
@@ -799,7 +909,39 @@ class CScript(bytes):
         """
         if checksize and len(self) > MAX_SCRIPT_ELEMENT_SIZE:
             raise ValueError("redeemScript exceeds max allowed size; P2SH output would be unspendable")
-        return CScript([OP_HASH160, bitcointx.core.Hash160(self), OP_EQUAL])
+        return self.__class__([OP_HASH160, bitcointx.core.Hash160(self), OP_EQUAL])
+
+    def to_p2wsh_scriptPubKey(self, checksize=True):
+        """Create P2WSH scriptPubKey from this redeemScript
+
+        That is, create the P2WSH scriptPubKey that requires this script as a
+        redeemScript to spend.
+
+        checksize - Check if the redeemScript is larger than the 520-byte max
+        pushdata limit; raise ValueError if limit exceeded.
+
+        Since a >520-byte PUSHDATA makes EvalScript() fail, it's not actually
+        possible to redeem P2WSH outputs with redeem scripts >520 bytes.
+        """
+        if checksize and len(self) > MAX_SCRIPT_ELEMENT_SIZE:
+            raise ValueError("redeemScript exceeds max allowed size; P2SH output would be unspendable")
+        return self.__class__([0, hashlib.sha256(self).digest()])
+
+    def to_p2wpkh_scriptPubKey(self, checksize=True):
+        """Create P2WPKH scriptPubKey from this redeemScript
+
+        That is, create the P2WPKH scriptPubKey that requires this script as a
+        redeemScript to spend.
+
+        checksize - Check if the redeemScript is larger than the 520-byte max
+        pushdata limit; raise ValueError if limit exceeded.
+
+        Since a >520-byte PUSHDATA makes EvalScript() fail, it's not actually
+        possible to redeem P2WSH outputs with redeem scripts >520 bytes.
+        """
+        if checksize and len(self) > MAX_SCRIPT_ELEMENT_SIZE:
+            raise ValueError("redeemScript exceeds max allowed size; P2WPKH output would be unspendable")
+        return self.__class__([0, bitcointx.core.Hash160(self)])
 
     def GetSigOpCount(self, fAccurate):
         """Get the SigOp count.
@@ -821,6 +963,30 @@ class CScript(bytes):
             lastOpcode = opcode
         return n
 
+    def sighash(self, txTo, inIdx, hashtype, amount=0, sigversion=SIGVERSION_BASE):
+        """Calculate a signature hash
+
+        'Cooked' version that checks if inIdx is out of bounds - this is *not*
+        consensus-correct behavior, but is what you probably want for general
+        wallet use.
+        """
+        (h, err) = self.raw_sighash(txTo, inIdx, hashtype, amount=amount, sigversion=sigversion)
+        if err is not None:
+            raise ValueError(err)
+        return h
+
+    def raw_sighash(self, txTo, inIdx, hashtype, amount=0, sigversion=SIGVERSION_BASE):
+        """Consensus-correct SignatureHash
+
+        Returns (hash, err) to precisely match the consensus-critical behavior of
+        the SIGHASH_SINGLE bug. (inIdx is *not* checked for validity)
+
+        Default implementation calls bitcoin-specific sighash function.
+
+        If you're just writing wallet software you probably want sighash() method instead."""
+        return RawBitcoinSignatureHash(self, txTo, inIdx, hashtype,
+                                       amount=amount, sigversion=sigversion)
+
 
 class CScriptWitness(ImmutableSerializable):
     """An encoding of the data elements on the initial stack for (segregated
@@ -829,7 +995,13 @@ class CScriptWitness(ImmutableSerializable):
     __slots__ = ['stack']
 
     def __init__(self, stack=()):
-        object.__setattr__(self, 'stack', stack)
+        coerced_stack = []
+        for (opcode, data, sop_idx) in CScript(stack).raw_iter():
+            if data is not None:
+                coerced_stack.append(data)
+            else:
+                coerced_stack.append(bytes([opcode]))
+        object.__setattr__(self, 'stack', tuple(coerced_stack))
 
     def __len__(self):
         return len(self.stack)
@@ -840,6 +1012,7 @@ class CScriptWitness(ImmutableSerializable):
     def __repr__(self):
         return 'CScriptWitness([' + ','.join("x('%s')" % bitcointx.core.b2x(s) for s in self.stack) + '])'
 
+    @no_bool_use_as_property
     def is_null(self):
         return len(self.stack) == 0
 
@@ -876,7 +1049,7 @@ def FindAndDelete(script, sig):
             skip = False
     if not skip:
         r += script[last_sop_idx:]
-    return CScript(r)
+    return script.__class__(r)
 
 
 def IsLowDERSignature(sig):
@@ -932,7 +1105,7 @@ def CompareBigEndian(c1, c2):
     return 0
 
 
-def RawSignatureHash(script, txTo, inIdx, hashtype, amount=0, sigversion=SIGVERSION_BASE):
+def RawBitcoinSignatureHash(script, txTo, inIdx, hashtype, amount=0, sigversion=SIGVERSION_BASE):
     """Consensus-correct SignatureHash
 
     Returns (hash, err) to precisely match the consensus-critical behavior of
@@ -941,6 +1114,8 @@ def RawSignatureHash(script, txTo, inIdx, hashtype, amount=0, sigversion=SIGVERS
     If you're just writing wallet software you probably want SignatureHash()
     instead.
     """
+    assert sigversion in (SIGVERSION_BASE, SIGVERSION_WITNESS_V0)
+
     HASH_ONE = b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
 
     if sigversion == SIGVERSION_WITNESS_V0:
@@ -989,11 +1164,14 @@ def RawSignatureHash(script, txTo, inIdx, hashtype, amount=0, sigversion=SIGVERS
 
     if inIdx >= len(txTo.vin):
         return (HASH_ONE, "inIdx %d out of range (%d)" % (inIdx, len(txTo.vin)))
-    txtmp = bitcointx.core.CMutableTransaction.from_tx(txTo)
+
+    txtmp = txTo.to_mutable()
 
     for txin in txtmp.vin:
         txin.scriptSig = b''
-    txtmp.vin[inIdx].scriptSig = FindAndDelete(script, CScript([OP_CODESEPARATOR]))
+
+    txtmp.vin[inIdx].scriptSig = FindAndDelete(
+        script, script.__class__([OP_CODESEPARATOR]))
 
     if (hashtype & 0x1f) == SIGHASH_NONE:
         txtmp.vout = []
@@ -1010,7 +1188,7 @@ def RawSignatureHash(script, txTo, inIdx, hashtype, amount=0, sigversion=SIGVERS
         tmp = txtmp.vout[outIdx]
         txtmp.vout = []
         for i in range(outIdx):
-            txtmp.vout.append(bitcointx.core.CTxOut())
+            txtmp.vout.append(bitcointx.core.CMutableTxOut())
         txtmp.vout.append(tmp)
 
         for i in range(len(txtmp.vin)):
@@ -1022,8 +1200,8 @@ def RawSignatureHash(script, txTo, inIdx, hashtype, amount=0, sigversion=SIGVERS
         txtmp.vin = []
         txtmp.vin.append(tmp)
 
-    txtmp.wit = bitcointx.core.CTxWitness()
-    s = txtmp.serialize()
+    txtmp.wit = bitcointx.core.CMutableTxWitness()
+    s = txtmp.serialize(for_sighash=True)
     s += struct.pack(b"<i", hashtype)
 
     hash = bitcointx.core.Hash(s)
@@ -1031,6 +1209,18 @@ def RawSignatureHash(script, txTo, inIdx, hashtype, amount=0, sigversion=SIGVERS
     return (hash, None)
 
 
+def RawSignatureHash(script, txTo, inIdx, hashtype, amount=0, sigversion=SIGVERSION_BASE):
+    """Consensus-correct SignatureHash
+
+    Returns (hash, err) to precisely match the consensus-critical behavior of
+    the SIGHASH_SINGLE bug. (inIdx is *not* checked for validity)
+
+    If you're just writing wallet software you probably want SignatureHash()
+    instead.
+    """
+    return script.raw_sighash(txTo, inIdx, hashtype, amount=amount, sigversion=sigversion)
+
+
 def SignatureHash(script, txTo, inIdx, hashtype, amount=0, sigversion=SIGVERSION_BASE):
     """Calculate a signature hash
 
@@ -1044,149 +1234,286 @@ def SignatureHash(script, txTo, inIdx, hashtype, amount=0, sigversion=SIGVERSION
     return h
 
 
+class CBitcoinScript(CScript, ScriptBitcoinClass):
+    ...
+
+
+def parse_standard_multisig_redeem_script(script):
+    """parse multisig script, raise ValueError if it does not match the
+    format of the script produced by construct_multisig_redeem_script"""
+    si = iter(script)
+    try:
+        required = next(si)
+    except StopIteration:
+        raise ValueError(
+            'script is too short: stoped at required number of signatures')
+    if not isinstance(required, int):
+        raise ValueError('script is not a p2sh script, required number '
+                         'of signatures is not a number')
+    if required < 1:
+        raise ValueError('required number of signatures are less than 1')
+    if required > MAX_P2SH_MULTISIG_PUBKEYS:
+        raise ValueError('required number of signatures are more than {}'
+                         .format(MAX_P2SH_MULTISIG_PUBKEYS))
+    total = None
+    pubkeys = []
+    for i in range(MAX_P2SH_MULTISIG_PUBKEYS+1):  # +1 for `total`
+        try:
+            pub = next(si)
+            if isinstance(pub, int):
+                total = pub  # pubkeys ended
+                break
+            assert isinstance(pub, bytes)
+            pubkeys.append(pub)
+        except StopIteration:
+            raise ValueError(
+                'script is too short for specified number of required '
+                'signatures ({})'.format(required))
+
+    if len(pubkeys) > MAX_P2SH_MULTISIG_PUBKEYS:
+        raise ValueError('script contains more than {} pubkeys'
+                         .format(MAX_P2SH_MULTISIG_PUBKEYS))
+
+    if total != len(pubkeys):
+        raise ValueError('number of pubkeys in the script is not the same '
+                         'as the number of pubkeys required')
+    if total < 2:
+        raise ValueError('total number of pubkeys are less than 2')
+    if total > MAX_P2SH_MULTISIG_PUBKEYS:
+        raise ValueError('required number of pubkeys are more than {}'
+                         .format(MAX_P2SH_MULTISIG_PUBKEYS))
+    try:
+        last_op = next(si)
+    except StopIteration:
+            raise ValueError('script is too short, ended before we gan get '
+                             'last opcode')
+
+    if last_op != OP_CHECKMULTISIG:
+        raise ValueError('script is not a p2sh script, last opcode '
+                         'is not OP_CHECKMULTISIG')
+
+    return {'total': total, 'required': required, 'pubkeys': pubkeys}
+
+
+def standard_multisig_witness_stack(sigs, redeem_script):
+
+    # check valid p2sh script
+    info = parse_standard_multisig_redeem_script(redeem_script)
+
+    if not all(isinstance(s, (bytes, bytearray)) for s in sigs):
+        raise ValueError('sigs must be an array of bytes (or bytearrays)')
+
+    if len(sigs) > info['total']:
+        raise ValueError('number of signatures ({}) is greater than '
+                         'total pubkeys ({}) in the redeem script'
+                         .format(len(sigs), info['total']))
+
+    if len(sigs) != info['required']:
+        raise ValueError('number of signatures ({}) does not match '
+                         'the number of required pubkeys ({}) '
+                         'in the redeem script'
+                         .format(len(sigs), info['required']))
+
+    stack = [0]  # dummy 0 required for CHECKMULTISIG
+    stack.extend(sigs)
+    stack.append(redeem_script)
+    return stack
+
+
+def standard_multisig_redeem_script(*, total=None, required=None, pubkeys=None):
+    """Construct multisignature redeem script.
+    We require to supply total number of pubkeys as separate argument
+    to be able to catch bugs when pubkeys array is wrong for some reason.
+    If the callers do not care about the possibility of such bug, they
+    can just supply total=len(pubkeys).
+
+    For '1-of-1' case the function raise ValueError exception,
+    because this case is ambiguous - can be done via CHECKSIG or CHECKMULTISIG,
+    and is not really a multisig.
+
+    Arguments:
+
+    total    - total number of pubkeys (must match the length of pubkeys array)
+    required - the number of signatures required to satisfy the script,
+               must be less than or equal to `total`
+    pubkeys  - an array of pubkeys"""
+
+    if total != len(pubkeys):
+        raise ValueError("'total' argument must match length of pubkeys array")
+    if required <= 0:
+        raise ValueError("'required' argument must be >= 1")
+    if required > total:
+        raise ValueError(
+            "'required' argument must be less than or equal to 'total'")
+    if total > MAX_P2SH_MULTISIG_PUBKEYS:
+        raise ValueError("%d pubkeys do not fit into standard p2sh multisig"
+                         % total)
+    if total == 1:
+        # for 1-of-1, there is ambiguity in possible scripts:
+        # using CHECKMULTISIG, or using just CHECKSIG.
+        # scriptSig for these variants would differ, because CHECKMULTISIG
+        # requires extra '0' on the stack, which CHECKSIG does not require.
+        # therefore, if the callers want to support 1-of-1 in P2SH,
+        # they need to handle it themselves.
+        raise ValueError('1-of-1 multisig is not supported')
+
+    return CScript([required] + pubkeys + [total, OP_CHECKMULTISIG])
+
+
+# default dispatcher for the module
+activate_class_dispatcher(ScriptBitcoinClassDispatcher)
+
+
 __all__ = (
-        'MAX_SCRIPT_SIZE',
-        'MAX_SCRIPT_ELEMENT_SIZE',
-        'MAX_SCRIPT_OPCODES',
-        'OPCODE_NAMES',
-        'CScriptOp',
+    'MAX_SCRIPT_SIZE',
+    'MAX_SCRIPT_ELEMENT_SIZE',
+    'MAX_SCRIPT_OPCODES',
+    'OPCODE_NAMES',
+    'CScriptOp',
 
-        # every opcode
-        'OP_0',
-        'OP_FALSE',
-        'OP_PUSHDATA1',
-        'OP_PUSHDATA2',
-        'OP_PUSHDATA4',
-        'OP_1NEGATE',
-        'OP_RESERVED',
-        'OP_1',
-        'OP_TRUE',
-        'OP_2',
-        'OP_3',
-        'OP_4',
-        'OP_5',
-        'OP_6',
-        'OP_7',
-        'OP_8',
-        'OP_9',
-        'OP_10',
-        'OP_11',
-        'OP_12',
-        'OP_13',
-        'OP_14',
-        'OP_15',
-        'OP_16',
-        'OP_NOP',
-        'OP_VER',
-        'OP_IF',
-        'OP_NOTIF',
-        'OP_VERIF',
-        'OP_VERNOTIF',
-        'OP_ELSE',
-        'OP_ENDIF',
-        'OP_VERIFY',
-        'OP_RETURN',
-        'OP_TOALTSTACK',
-        'OP_FROMALTSTACK',
-        'OP_2DROP',
-        'OP_2DUP',
-        'OP_3DUP',
-        'OP_2OVER',
-        'OP_2ROT',
-        'OP_2SWAP',
-        'OP_IFDUP',
-        'OP_DEPTH',
-        'OP_DROP',
-        'OP_DUP',
-        'OP_NIP',
-        'OP_OVER',
-        'OP_PICK',
-        'OP_ROLL',
-        'OP_ROT',
-        'OP_SWAP',
-        'OP_TUCK',
-        'OP_CAT',
-        'OP_SUBSTR',
-        'OP_LEFT',
-        'OP_RIGHT',
-        'OP_SIZE',
-        'OP_INVERT',
-        'OP_AND',
-        'OP_OR',
-        'OP_XOR',
-        'OP_EQUAL',
-        'OP_EQUALVERIFY',
-        'OP_RESERVED1',
-        'OP_RESERVED2',
-        'OP_1ADD',
-        'OP_1SUB',
-        'OP_2MUL',
-        'OP_2DIV',
-        'OP_NEGATE',
-        'OP_ABS',
-        'OP_NOT',
-        'OP_0NOTEQUAL',
-        'OP_ADD',
-        'OP_SUB',
-        'OP_MUL',
-        'OP_DIV',
-        'OP_MOD',
-        'OP_LSHIFT',
-        'OP_RSHIFT',
-        'OP_BOOLAND',
-        'OP_BOOLOR',
-        'OP_NUMEQUAL',
-        'OP_NUMEQUALVERIFY',
-        'OP_NUMNOTEQUAL',
-        'OP_LESSTHAN',
-        'OP_GREATERTHAN',
-        'OP_LESSTHANOREQUAL',
-        'OP_GREATERTHANOREQUAL',
-        'OP_MIN',
-        'OP_MAX',
-        'OP_WITHIN',
-        'OP_RIPEMD160',
-        'OP_SHA1',
-        'OP_SHA256',
-        'OP_HASH160',
-        'OP_HASH256',
-        'OP_CODESEPARATOR',
-        'OP_CHECKSIG',
-        'OP_CHECKSIGVERIFY',
-        'OP_CHECKMULTISIG',
-        'OP_CHECKMULTISIGVERIFY',
-        'OP_NOP1',
-        'OP_NOP2',
-        'OP_CHECKLOCKTIMEVERIFY',
-        'OP_NOP3',
-        'OP_NOP4',
-        'OP_NOP5',
-        'OP_NOP6',
-        'OP_NOP7',
-        'OP_NOP8',
-        'OP_NOP9',
-        'OP_NOP10',
-        'OP_SMALLINTEGER',
-        'OP_PUBKEYS',
-        'OP_PUBKEYHASH',
-        'OP_PUBKEY',
-        'OP_INVALIDOPCODE',
+    # every opcode
+    'OP_0',
+    'OP_FALSE',
+    'OP_PUSHDATA1',
+    'OP_PUSHDATA2',
+    'OP_PUSHDATA4',
+    'OP_1NEGATE',
+    'OP_RESERVED',
+    'OP_1',
+    'OP_TRUE',
+    'OP_2',
+    'OP_3',
+    'OP_4',
+    'OP_5',
+    'OP_6',
+    'OP_7',
+    'OP_8',
+    'OP_9',
+    'OP_10',
+    'OP_11',
+    'OP_12',
+    'OP_13',
+    'OP_14',
+    'OP_15',
+    'OP_16',
+    'OP_NOP',
+    'OP_VER',
+    'OP_IF',
+    'OP_NOTIF',
+    'OP_VERIF',
+    'OP_VERNOTIF',
+    'OP_ELSE',
+    'OP_ENDIF',
+    'OP_VERIFY',
+    'OP_RETURN',
+    'OP_TOALTSTACK',
+    'OP_FROMALTSTACK',
+    'OP_2DROP',
+    'OP_2DUP',
+    'OP_3DUP',
+    'OP_2OVER',
+    'OP_2ROT',
+    'OP_2SWAP',
+    'OP_IFDUP',
+    'OP_DEPTH',
+    'OP_DROP',
+    'OP_DUP',
+    'OP_NIP',
+    'OP_OVER',
+    'OP_PICK',
+    'OP_ROLL',
+    'OP_ROT',
+    'OP_SWAP',
+    'OP_TUCK',
+    'OP_CAT',
+    'OP_SUBSTR',
+    'OP_LEFT',
+    'OP_RIGHT',
+    'OP_SIZE',
+    'OP_INVERT',
+    'OP_AND',
+    'OP_OR',
+    'OP_XOR',
+    'OP_EQUAL',
+    'OP_EQUALVERIFY',
+    'OP_RESERVED1',
+    'OP_RESERVED2',
+    'OP_1ADD',
+    'OP_1SUB',
+    'OP_2MUL',
+    'OP_2DIV',
+    'OP_NEGATE',
+    'OP_ABS',
+    'OP_NOT',
+    'OP_0NOTEQUAL',
+    'OP_ADD',
+    'OP_SUB',
+    'OP_MUL',
+    'OP_DIV',
+    'OP_MOD',
+    'OP_LSHIFT',
+    'OP_RSHIFT',
+    'OP_BOOLAND',
+    'OP_BOOLOR',
+    'OP_NUMEQUAL',
+    'OP_NUMEQUALVERIFY',
+    'OP_NUMNOTEQUAL',
+    'OP_LESSTHAN',
+    'OP_GREATERTHAN',
+    'OP_LESSTHANOREQUAL',
+    'OP_GREATERTHANOREQUAL',
+    'OP_MIN',
+    'OP_MAX',
+    'OP_WITHIN',
+    'OP_RIPEMD160',
+    'OP_SHA1',
+    'OP_SHA256',
+    'OP_HASH160',
+    'OP_HASH256',
+    'OP_CODESEPARATOR',
+    'OP_CHECKSIG',
+    'OP_CHECKSIGVERIFY',
+    'OP_CHECKMULTISIG',
+    'OP_CHECKMULTISIGVERIFY',
+    'OP_NOP1',
+    'OP_NOP2',
+    'OP_CHECKLOCKTIMEVERIFY',
+    'OP_NOP3',
+    'OP_NOP4',
+    'OP_NOP5',
+    'OP_NOP6',
+    'OP_NOP7',
+    'OP_NOP8',
+    'OP_NOP9',
+    'OP_NOP10',
+    'OP_SMALLINTEGER',
+    'OP_PUBKEYS',
+    'OP_PUBKEYHASH',
+    'OP_PUBKEY',
+    'OP_INVALIDOPCODE',
 
-        'OPCODES_BY_NAME',
-        'DISABLED_OPCODES',
-        'CScriptInvalidError',
-        'CScriptTruncatedPushDataError',
-        'CScript',
-        'CScriptWitness',
-        'SIGHASH_ALL',
-        'SIGHASH_NONE',
-        'SIGHASH_SINGLE',
-        'SIGHASH_ANYONECANPAY',
-        'FindAndDelete',
-        'RawSignatureHash',
-        'SignatureHash',
-        'IsLowDERSignature',
+    'OPCODES_BY_NAME',
+    'DISABLED_OPCODES',
+    'CScriptInvalidError',
+    'CScriptTruncatedPushDataError',
+    'CScript',
+    'CScriptWitness',
+    'SIGHASH_ALL',
+    'SIGHASH_NONE',
+    'SIGHASH_SINGLE',
+    'SIGHASH_ANYONECANPAY',
+    'FindAndDelete',
+    'RawSignatureHash',
+    'RawBitcoinSignatureHash',
+    'SignatureHash',
+    'IsLowDERSignature',
 
-        'SIGVERSION_BASE',
-        'SIGVERSION_WITNESS_V0',
+    'SIGVERSION_BASE',
+    'SIGVERSION_WITNESS_V0',
+
+    'ScriptCoinClassDispatcher',
+    'ScriptCoinClass',
+    'parse_standard_multisig_redeem_script',
+    'standard_multisig_redeem_script',
+    'standard_multisig_witness_stack',
 )
